package eu.jonahbauer.chat.bot.config;

import lombok.AccessLevel;
import lombok.EqualsAndHashCode;
import lombok.RequiredArgsConstructor;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.Unmodifiable;

import java.util.*;

@SuppressWarnings("unused")
@RequiredArgsConstructor(access = AccessLevel.PRIVATE)
@EqualsAndHashCode(onlyExplicitlyIncluded = true)
public final class BotConfig {
    public static final Key.OfString NAME = Key.ofString("name");
    public static final Key.OfBoolean BOT_TAG = Key.ofBoolean("bottag");
    public static final Key.OfBoolean IGNORE_BOTS = Key.ofBoolean("ignore_bots");
    public static final Key.OfBoolean PUBLIC_ID = Key.ofBoolean("public_id");
    public static final Key.OfStringArray CHANNELS = Key.ofStringArray("channels");
    public static final Key.OfString TYPE = Key.ofString("type");
    public static final BotConfig EMPTY = new BotConfig(Collections.emptyMap());

    @EqualsAndHashCode.Include
    private final @Unmodifiable @NotNull Map<@NotNull String, @NotNull Value> data;

    public static @NotNull Builder builder() {
        return new Builder();
    }

    /**
     * {@return a new builder initialized with this config's data}
     */
    public @NotNull Builder mutate() {
        return new Builder(this.data);
    }

    @Override
    public @NotNull String toString() {
        return "BotConfig" + data;
    }

    /**
     * Merges the other config into this one, returning a new config container properties of both this and the provided
     * config. Properties present in both this and the provided config will contain the value of the provided config.
     * @param config another config
     * @return a merged config
     */
    public @NotNull BotConfig merge(@NotNull BotConfig config) {
        var data = new HashMap<>(this.data);
        data.putAll(config.data);
        return new BotConfig(Map.copyOf(data));
    }

    /**
     * {@return the set of all keys to this config}
     */
    public @NotNull Set<@NotNull String> keySet() {
        return data.keySet();
    }

    private transient @Nullable Set<@NotNull Entry<?>> entrySet;
    public @NotNull Set<@NotNull Entry<?>> entrySet() {
        if (entrySet == null) {
            var entries = new HashSet<Entry<?>>();
            data.forEach((key, v) -> entries.add(switch (v) {
                case Value.OfLong(long value) -> Entry.ofLong(key, value);
                case Value.OfDouble(double value) -> Entry.ofDouble(key, value);
                case Value.OfString(String value) -> Entry.ofString(key, value);
                case Value.OfBoolean(boolean value) -> Entry.ofBoolean(key, value);
                case Value.OfStringArray(List<String> value) -> Entry.ofStringArray(key, value);
            }));
            entrySet = Collections.unmodifiableSet(entries);
        }
        return entrySet;
    }

    /*
     * getters for default keys
     */

    public @NotNull String getType() {
        return get(TYPE).orElseThrow(() -> new BotConfigurationException("missing property: " + TYPE));
    }

    public @NotNull String getName() {
        return get(NAME).orElseThrow(() -> new BotConfigurationException("missing property: " + NAME));
    }

    public @NotNull List<@NotNull String> getChannels() {
        return get(CHANNELS).orElseThrow(() -> new BotConfigurationException("missing property: " + CHANNELS));
    }

    public boolean isBotTag() {
        return get(BOT_TAG).orElse(true);
    }

    public boolean isIgnoreBots() {
        return get(IGNORE_BOTS).orElse(true);
    }

    public boolean isPublicId() {
        return get(PUBLIC_ID).orElse(true);
    }

    /*
     * generic getters
     */

    /**
     * {@return whether this config contains the given key}
     */
    public boolean has(@NotNull String key) {
        return data.containsKey(key);
    }

    /**
     * {@return whether this config contains the given key and the value associated with the key matches the key type}
     */
    public boolean has(@NotNull Key<?, ?> key) {
        return switch (key) {
            case Key.OfBoolean(var k) -> getBoolean(k).isPresent();
            case Key.OfLong(var k) -> getLong(k).isPresent();
            case Key.OfDouble(var k) -> getDouble(k).isPresent();
            case Key.OfStringArray(var k) -> getStringArray(k).isPresent();
            case Key.OfString(var k) -> getString(k).isPresent();
        };
    }

    /**
     * {@return the value associated with the given key}
     * @throws BotConfigurationException if no value is present for the given key or the value does not match the key
     * type
     */
    @SuppressWarnings("unchecked")
    public <T> @NotNull T require(@NotNull Key<T, ?> key) {
        try {
            return (T) switch (key) {
                case Key.OfBoolean(var k) -> getBoolean(k).orElseThrow();
                case Key.OfLong(var k) -> getLong(k).orElseThrow();
                case Key.OfDouble(var k) -> getDouble(k).orElseThrow();
                case Key.OfStringArray(var k) -> getStringArray(k).orElseThrow();
                case Key.OfString(var k) -> getString(k).orElseThrow();
            };
        } catch (NoSuchElementException ex) {
            throw new BotConfigurationException("Missing property value:" + key.name(), ex);
        }
    }

    /**
     * {@return the value associated with the given key} Returns an empty optional if no value is present for the given
     * key or the value does not match the key type.
     */
    @SuppressWarnings("unchecked")
    public <T> @NotNull T get(@NotNull Key<?, T> key) {
        return (T) switch (key) {
            case Key.OfBoolean(var k) -> getBoolean(k);
            case Key.OfLong(var k) -> getLong(k);
            case Key.OfDouble(var k) -> getDouble(k);
            case Key.OfStringArray(var k) -> getStringArray(k);
            case Key.OfString(var k) -> getString(k);
        };
    }

    /**
     * {@return the value associated with the given key} If not empty, then the returned {@link Optional} will
     * contain an object of type {@link String}, {@link Long}, {@link Double}, {@link Boolean} or {@link List List&lt;String&gt;}.
     */
    public @NotNull Optional<?> get(@NotNull String key) {
        var value = data.get(key);
        return value == null ? Optional.empty() : Optional.of(value.asObject());
    }

    /**
     * {@return the value associated with the given key} Returns an {@linkplain Optional#empty() empty optional} if no
     * value is present for the given key or the value is not of type {@link String}.
     */
    public @NotNull Optional<String> getString(@NotNull String key) {
        var value = data.get(key);
        return value == null ? Optional.empty() : Optional.of(value.asString(key));
    }

    /**
     * {@return the value associated with the given key} Returns an {@linkplain OptionalLong#empty() empty optional} if
     * no value is present for the given key or the value is not of type {@link Long}.
     */
    public @NotNull OptionalLong getLong(@NotNull String key) {
        var value = data.get(key);
        return value == null ? OptionalLong.empty() : OptionalLong.of(value.asLong(key));
    }

    /**
     * {@return the value associated with the given key} Returns an {@linkplain OptionalDouble#empty() empty optional}
     * if no value is present for the given key or the value is not of type {@link Double}.
     */
    public @NotNull OptionalDouble getDouble(@NotNull String key) {
        var value = data.get(key);
        return value == null ? OptionalDouble.empty() : OptionalDouble.of(value.asDouble(key));
    }

    /**
     * {@return the value associated with the given key} Returns an {@linkplain Optional#empty() empty optional} if no
     * value is present for the given key or the value is not of type {@link Boolean}.
     */
    public @NotNull Optional<Boolean> getBoolean(@NotNull String key) {
        var value = data.get(key);
        return value == null ? Optional.empty() : Optional.of(value.asBoolean(key));
    }

    /**
     * {@return the value associated with the given key} Returns an {@linkplain Optional#empty() empty optional} if no
     * value is present for the given key or the value is not of type {@link List List&lt;String&gt;}.
     */
    public @NotNull Optional<List<String>> getStringArray(@NotNull String key) {
        var value = data.get(key);
        return value == null ? Optional.empty() : Optional.of(value.asArray(key));
    }

    private static BotConfigurationException wrongType(@NotNull String key, @NotNull String expected, @NotNull String actual) {
        return new BotConfigurationException("Wrong property type for " + key + " (expected " + expected + ", got " + actual + ")");
    }

    public static class Builder {
        private final Map<String, Value> data;

        private Builder() {
            this(Collections.emptyMap());
        }

        private Builder(Map<String, Value> data) {
            this.data = new HashMap<>(data);
        }

        public Builder type(@NotNull String type) {
            return value(TYPE, type);
        }

        public Builder name(@NotNull String name) {
            return value(NAME, name);
        }

        public Builder channels(@NotNull String @NotNull... channels) {
            return value(CHANNELS, List.of(channels));
        }

        public Builder channels(@NotNull List<@NotNull String> channels) {
            return value(CHANNELS, List.copyOf(channels));
        }

        public Builder noPublicId() {
            return value(PUBLIC_ID, false);
        }

        public Builder noBotTag() {
            return value(BOT_TAG, false);
        }

        public Builder noIgnoreBots() {
            return value(IGNORE_BOTS, false);
        }

        public Builder value(@NotNull String key, long value) {
            data.put(key, new Value.OfLong(value));
            return this;
        }

        public Builder value(@NotNull String key, double value) {
            data.put(key, new Value.OfDouble(value));
            return this;
        }

        public Builder value(@NotNull String key, @NotNull String value) {
            data.put(key, new Value.OfString(value));
            return this;
        }

        public Builder value(@NotNull String key, boolean value) {
            data.put(key, new Value.OfBoolean(value));
            return this;
        }

        public Builder value(@NotNull String key, @NotNull List<@NotNull String> value) {
            data.put(key, new Value.OfStringArray(value));
            return this;
        }

        public <S> Builder value(@NotNull Key<S, ?> key, S value) {
            data.put(key.name(), Value.forKey(key, value));
            return this;
        }

        public BotConfig build() {
            return new BotConfig(Map.copyOf(data));
        }
    }

    /**
     * A typed key to a {@link BotConfig}.
     * @param <S> the raw value type
     * @param <T> the optional value type
     */
    public sealed interface Key<S, T> {
        @NotNull String name();

        static @NotNull Key.OfLong ofLong(@NotNull String key) {
            return new OfLong(key);
        }

        static @NotNull Key.OfDouble ofDouble(@NotNull String key) {
            return new OfDouble(key);
        }

        static @NotNull Key.OfString ofString(@NotNull String key) {
            return new OfString(key);
        }

        static @NotNull Key.OfBoolean ofBoolean(@NotNull String key) {
            return new OfBoolean(key);
        }

        static @NotNull Key.OfStringArray ofStringArray(@NotNull String key) {
            return new OfStringArray(key);
        }

        record OfLong(@NotNull String name) implements Key<Long, OptionalLong> {}
        record OfDouble(@NotNull String name) implements Key<Double, OptionalDouble> {}
        record OfString(@NotNull String name) implements Key<String, Optional<String>> {}
        record OfBoolean(@NotNull String name) implements Key<Boolean, Optional<Boolean>> {}
        record OfStringArray(@NotNull String name) implements Key<List<String>, Optional<List<String>>> {}
    }

    public sealed interface Entry<T> extends Map.Entry<String, T> {
        @NotNull String key();
        @NotNull T value();

        @Override
        default @NotNull String getKey() {
            return key();
        }

        @Override
        default @NotNull T getValue() {
            return value();
        }

        @Override
        default T setValue(T value) {
            throw new UnsupportedOperationException();
        }

        static @NotNull Entry.OfLong ofLong(@NotNull String key, long value) {
            return new Entry.OfLong(key, value);
        }

        static @NotNull Entry.OfDouble ofDouble(@NotNull String key, double value) {
            return new Entry.OfDouble(key, value);
        }

        static @NotNull Entry.OfString ofString(@NotNull String key, @NotNull String value) {
            return new Entry.OfString(key, value);
        }

        static @NotNull Entry.OfBoolean ofBoolean(@NotNull String key, boolean value) {
            return new Entry.OfBoolean(key, value);
        }

        static @NotNull Entry.OfStringArray ofStringArray(@NotNull String key, @NotNull List<@NotNull String> value) {
            return new Entry.OfStringArray(key, value);
        }


        record OfLong(@NotNull String key, long valueAsLong) implements Entry<Long> {
            @Override
            public @NotNull Long value() {
                return valueAsLong;
            }
        }
        record OfDouble(@NotNull String key, double valueAsDouble) implements Entry<Double> {
            @Override
            public @NotNull Double value() {
                return valueAsDouble;
            }
        }
        record OfString(@NotNull String key, @NotNull String value) implements Entry<String> {}
        record OfBoolean(@NotNull String key, boolean valueAsBoolean) implements Entry<Boolean> {
            @Override
            public @NotNull Boolean value() {
                return valueAsBoolean;
            }
        }
        record OfStringArray(@NotNull String key, @NotNull List<@NotNull String> value) implements Entry<List<String>> {}
    }

    private sealed interface Value {
        String name();

        default long asLong(@NotNull String key) { throw wrongType(key, "integer", name()); }
        default double asDouble(@NotNull String key) { throw wrongType(key, "float", name()); }
        default String asString(@NotNull String key) { throw wrongType(key, "string", name()); }
        default boolean asBoolean(@NotNull String key) { throw wrongType(key, "boolean", name()); }
        default List<String> asArray(@NotNull String key) { throw wrongType(key, "array", name()); }
        Object asObject();

        @SuppressWarnings("unchecked")
        static <S> Value forKey(Key<S,?> key, S value) {
            return switch (key) {
                case Key.OfString _ -> new Value.OfString((String) value);
                case Key.OfDouble _ -> new Value.OfDouble((double) value);
                case Key.OfLong _ -> new Value.OfLong((long) value);
                case Key.OfBoolean _ -> new Value.OfBoolean((boolean) value);
                case Key.OfStringArray _ -> new Value.OfStringArray((List<String>) value);
            };
        }

        record OfLong(long value) implements Value {
            public String name() { return "integer"; }

            @Override
            public long asLong(@NotNull String key) { return value; }

            @Override
            public Object asObject() { return value; }
        }
        record OfDouble(double value) implements Value {
            public String name() { return "float"; }

            @Override
            public double asDouble(@NotNull String key) { return value; }

            @Override
            public Object asObject() { return value; }
        }
        record OfString(@NotNull String value) implements Value {
            public OfString { Objects.requireNonNull(value); }

            public String name() { return "string"; }

            @Override
            public String asString(@NotNull String key) { return value; }

            @Override
            public Object asObject() { return value; }
        }
        record OfBoolean(boolean value) implements Value {
            public String name() { return "boolean"; }

            @Override
            public boolean asBoolean(@NotNull String key) { return value; }

            @Override
            public Object asObject() { return value; }
        }
        record OfStringArray(@NotNull List<@NotNull String> value) implements Value {
            public OfStringArray { value = List.copyOf(value); }
            public String name() { return "array"; }

            @Override
            public List<String> asArray(@NotNull String key) { return value; }

            @Override
            public Object asObject() { return value; }
        }
    }
}
