package eu.jonahbauer.chat.bot.database;

import com.zaxxer.hikari.HikariConfig;
import com.zaxxer.hikari.HikariDataSource;
import eu.jonahbauer.chat.bot.database.impl.Binder;
import eu.jonahbauer.chat.bot.database.impl.Helper;
import eu.jonahbauer.chat.bot.database.impl.Inserter;
import eu.jonahbauer.chat.bot.database.impl.Updater;
import eu.jonahbauer.chat.bot.database.impl.types.TypeAdapter;
import lombok.Lombok;
import lombok.SneakyThrows;
import org.intellij.lang.annotations.Language;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;

@SuppressWarnings({"SqlSourceToSinkFlow", "unused"})
public class Database implements AutoCloseable {
    private static final ScopedValue<Connection> CONNECTION = ScopedValue.newInstance();
    private final @NotNull HikariDataSource ds;

    public Database(@NotNull String url) {
        this(url, null, null);
    }

    public Database(@NotNull String url, @Nullable String user, @Nullable String password) {
        var hikariConfig = new HikariConfig();
        hikariConfig.setJdbcUrl(url);
        hikariConfig.setUsername(user);
        hikariConfig.setPassword(password);
        hikariConfig.addDataSourceProperty("enforceForeignKeys", true);
        this.ds = new HikariDataSource(hikariConfig);
    }


    @SuppressWarnings("RedundantThrows")
    public <T extends Throwable> void transaction(@NotNull ThrowingRunnable<T> runnable) throws SQLException, T {
        if (CONNECTION.isBound()) throw new IllegalStateException("nested transactions are not supported");

        try (var connection = getConnection0()) {
            connection.setAutoCommit(false);
            try {
                ScopedValue.runWhere(CONNECTION, connection, () -> {
                    try {
                        runnable.run();
                    } catch (Throwable t) {
                        throw Lombok.sneakyThrow(t);
                    }
                });
                connection.commit();
            } catch (Throwable t) {
                connection.rollback();
                throw t;
            }
        }
    }

    @SuppressWarnings("RedundantThrows")
    public <R, T extends Throwable> R transaction(@NotNull ThrowingCallable<R, T> callable) throws SQLException, T {
        if (CONNECTION.isBound()) throw new IllegalStateException("nested transactions are not supported");

        try (var connection = getConnection0()) {
            connection.setAutoCommit(false);
            try {
                var result = ScopedValue.callWhere(CONNECTION, connection, () -> {
                    try {
                        return callable.call();
                    } catch (Throwable t) {
                        throw Lombok.sneakyThrow(t);
                    }
                });
                connection.commit();
                return result;
            } catch (Throwable t) {
                connection.rollback();
                throw Lombok.sneakyThrow(t);
            }
        }
    }

    public <T extends Record> @NotNull List<T> findAll(@NotNull Class<T> type) throws SQLException {
        return executeQuery(type, "SELECT * FROM " + Helper.getQuotedTableName(type));
    }

    public <T extends Record & Entity<T>> @NotNull Optional<T> findById(@NotNull Class<T> type, long id) throws SQLException {
        return findWhere(type, "id", id);
    }

    @SuppressWarnings("Convert2MethodRef")
    public <T extends Record> @NotNull List<T> findAllWhere(@NotNull Class<T> type, @NotNull String component, Object value) throws SQLException {
        return findWhere0(type, component, value, (t, query, customizer) -> executeQuery(t, query, customizer));
    }

    @SuppressWarnings("Convert2MethodRef")
    public <T extends Record> @NotNull Optional<T> findWhere(@NotNull Class<T> type, @NotNull String component, Object value) throws SQLException {
        return findWhere0(type, component, value, (t, query, customizer) -> executeUniqueQuery(t, query, customizer));
    }

    @SneakyThrows(NoSuchMethodException.class)
    private <T extends Record, R> R findWhere0(@NotNull Class<T> type, @NotNull String component, Object value, @NotNull FindWhere<T, R> finder) throws SQLException {
        var table = Helper.getQuotedTableName(type);
        var column = Helper.getColumnName(type, component);
        var componentType = type.getMethod(component).getReturnType();
        var adapter = TypeAdapter.get(componentType);
        return finder.find(
                type,
                "SELECT * FROM " + table + " WHERE " + column + " " + (value == null ? "IS NULL" : "= ?"),
                stmt -> {
                    if (value != null) adapter.prepareUnsafe(stmt, 1, value);
                }
        );
    }

    private interface FindWhere<T extends Record, R> {
        R find(@NotNull Class<T> type, @Language("SQL") @NotNull String query, @NotNull ThrowingConsumer<PreparedStatement, SQLException> customizer) throws SQLException;
    }

    public <T extends Record> @NotNull List<T> bindAll(@NotNull Class<T> type, @NotNull ResultSet result) throws SQLException {
        return bindAll(type, result, "");
    }

    public <T extends Record> @NotNull List<T> bindAll(@NotNull Class<T> type, @NotNull ResultSet result, @NotNull String prefix) throws SQLException {
        var out = new ArrayList<T>();
        var binder = Binder.get(type);
        while (result.next()) {
            out.add(binder.bind(result, prefix));
        }
        return Collections.unmodifiableList(out);
    }

    public <T extends Record> @NotNull Optional<T> bind(@NotNull Class<T> type, @NotNull ResultSet result) throws SQLException {
        return bind(type, result, "");
    }

    public <T extends Record> @NotNull Optional<T> bind(@NotNull Class<T> type, @NotNull ResultSet result, @NotNull String prefix) throws SQLException {
        if (result.next()) {
            var binder = Binder.get(type);
            var out = Optional.of(binder.bind(result, prefix));
            if (result.next()) throw new NonUniqueResultException();
            return out;
        } else {
            return Optional.empty();
        }
    }

    @SuppressWarnings("unchecked")
    public <T extends Record & Entity<T>> boolean delete(@NotNull T entity) throws SQLException {
        return delete((Class<T>) entity.getClass(), entity.id());
    }

    public <T extends Record & Entity<T>> boolean delete(@NotNull Class<T> type, long id) throws SQLException {
        return 1 == executeUpdate("DELETE FROM " + Helper.getQuotedTableName(type) + " WHERE id = ?", stmt -> stmt.setLong(1, id));
    }

    public long insert(@NotNull Record record) throws SQLException {
        return insert0(record);
    }

    private <T extends Record> long insert0(@NotNull T record) throws SQLException {
        @SuppressWarnings("unchecked")
        var type = (Class<T>) record.getClass();
        var connection = getConnection0();
        try {
            return Inserter.get(type).insert(connection, record);
        } finally {
            close(connection);
        }
    }

    public <T extends Record> T save(@NotNull T record) throws SQLException {
        @SuppressWarnings("unchecked")
        var type = (Class<T>) record.getClass();
        var connection = getConnection0();
        try {
            var id = Inserter.get(type).insert(connection, record);
            if (record instanceof Entity<?> entity) {
                return (T) entity.withId(id);
            } else {
                return record;
            }
        } finally {
            close(connection);
        }
    }

    public <T extends Record & Entity<T>> boolean update(@NotNull T entity) throws SQLException {
        @SuppressWarnings("unchecked")
        var type = (Class<T>) entity.getClass();
        var connection = getConnection0();
        try {
            return Updater.get(type).update(connection, entity);
        } finally {
            close(connection);
        }
    }

    @SuppressWarnings("SqlWithoutWhere")
    public int truncate(@NotNull Class<? extends Record> type) throws SQLException {
        return executeUpdate("DELETE FROM " + Helper.getQuotedTableName(type));
    }


    public <T extends Record> @NotNull List<T> executeQuery(@NotNull Class<T> type, @NotNull @Language("SQL") String query) throws SQLException {
        var connection = getConnection0();
        try (var stmt = connection.createStatement()) {
            Helper.logSqlStatement(query);
            try (var result = stmt.executeQuery(query)) {
                return bindAll(type, result);
            }
        } finally {
            close(connection);
        }
    }

    public <R extends Record, T extends Throwable> @NotNull List<R> executeQuery(@NotNull Class<R> type, @NotNull @Language("SQL") String query, @NotNull ThrowingConsumer<@NotNull PreparedStatement, T> customizer) throws SQLException, T {
        var connection = getConnection0();
        Helper.logSqlStatement(query);
        try (var stmt = connection.prepareStatement(query)) {
            customizer.accept(stmt);
            try (var result = stmt.executeQuery()) {
                return bindAll(type, result);
            }
        } finally {
            close(connection);
        }
    }

    public <T extends Record> @NotNull Optional<T> executeUniqueQuery(@NotNull Class<T> type, @NotNull @Language("SQL") String query) throws SQLException {
        var connection = getConnection0();
        try (var stmt = connection.createStatement()) {
            Helper.logSqlStatement(query);
            try (var result = stmt.executeQuery(query)) {
                return bind(type, result);
            }
        } finally {
            close(connection);
        }
    }

    public <R extends Record, T extends Throwable> @NotNull Optional<R> executeUniqueQuery(@NotNull Class<R> type, @NotNull @Language("SQL") String query, @NotNull ThrowingConsumer<@NotNull PreparedStatement, T> customizer) throws SQLException, T {
        var connection = getConnection0();
        Helper.logSqlStatement(query);
        try (var stmt = connection.prepareStatement(query)) {
            customizer.accept(stmt);
            try (var result = stmt.executeQuery()) {
                return bind(type, result);
            }
        } finally {
            close(connection);
        }
    }

    public int executeUpdate(@NotNull @Language("SQL") String query) throws SQLException {
        var connection = getConnection0();
        try (var stmt = connection.createStatement()) {
            Helper.logSqlStatement(query);
            return stmt.executeUpdate(query);
        } finally {
            close(connection);
        }
    }

    public <T extends Throwable> int executeUpdate(@NotNull @Language("SQL") String query, @NotNull ThrowingConsumer<@NotNull PreparedStatement, T> customizer) throws SQLException, T {
        var connection = getConnection0();
        Helper.logSqlStatement(query);
        try (var stmt = connection.prepareStatement(query)) {
            customizer.accept(stmt);
            return stmt.executeUpdate();
        } finally {
            close(connection);
        }
    }

    public @NotNull Connection getConnection() throws SQLException {
        return ds.getConnection();
    }

    private @NotNull Connection getConnection0() throws SQLException {
        return CONNECTION.isBound() ? CONNECTION.get() : ds.getConnection();
    }

    void close(@NotNull Connection connection) throws SQLException {
        if (!CONNECTION.isBound() || CONNECTION.get() != connection) connection.close();
    }

    @Override
    public void close() {
        ds.close();
    }

    public interface ThrowingCallable<R, T extends Throwable> {
        R call() throws T;
    }

    public interface ThrowingRunnable<T extends Throwable> {
        void run() throws T;
    }

    public interface ThrowingConsumer<S, T extends Throwable> {
        void accept(S object) throws T;
    }
}
