package eu.jonahbauer.chat.bot.database.impl;

import eu.jonahbauer.chat.bot.database.impl.types.TypeAdapter;
import lombok.Lombok;
import org.jetbrains.annotations.NotNull;

import java.lang.reflect.Method;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

final class PreparerFactory extends AbstractFactory<Preparer<?>> {
    public static final @NotNull PreparerFactory INSTANCE = new PreparerFactory();

    private PreparerFactory() {}

    @Override
    protected  <T extends Record> @NotNull Preparer<T> computeValue0(@NotNull Class<T> type) {
        var components = type.getRecordComponents();

        var columns = new ArrayList<String>(components.length - 1);
        var accessors = new ArrayList<Method>(components.length - 1);
        var adapters = new ArrayList<TypeAdapter<?>>(components.length - 1);

        for (var component : components) {
            if (Helper.isIdColumn(component)) continue;

            columns.add(Helper.getQuotedColumnName(component));
            accessors.add(component.getAccessor());
            adapters.add(TypeAdapter.get(component.getType()));
        }

        return new PreparerImpl<>(columns, adapters, accessors);
    }

    private static final class PreparerImpl<T extends Record> implements Preparer<T> {
        private final @NotNull List<@NotNull String> columns;
        private final @NotNull List<@NotNull TypeAdapter<?>> adapters;
        private final @NotNull List<@NotNull Method> accessors;

        public PreparerImpl(
                @NotNull List<@NotNull String> columns,
                @NotNull List<@NotNull TypeAdapter<?>> adapters,
                @NotNull List<@NotNull Method> accessors
        ) {
            this.columns = List.copyOf(columns);
            this.adapters = List.copyOf(adapters);
            this.accessors = List.copyOf(accessors);
            if (this.columns.size() != this.adapters.size() || this.adapters.size() != this.accessors.size()) {
                throw new IllegalArgumentException();
            }
        }

        @Override
        public void prepare(@NotNull PreparedStatement stmt, @NotNull T entity, int offset) throws SQLException {
            try {
                for (int i = 0, length = columns.size(); i < length; i++) {
                    var value = accessors.get(i).invoke(entity);
                    adapters.get(i).prepareUnsafe(stmt, i + offset, value);
                }
            } catch (SQLException ex) {
                throw ex;
            } catch (Exception ex) {
                throw Lombok.sneakyThrow(ex);
            }
        }

        @Override
        public @NotNull List<@NotNull String> columns() {
            return columns;
        }
    }
}
