package eu.jonahbauer.chat.bot.pizza;

import eu.jonahbauer.chat.bot.api.ChatBot;
import eu.jonahbauer.chat.bot.api.Message.Post;
import eu.jonahbauer.chat.bot.config.BotConfig;
import eu.jonahbauer.chat.bot.database.Database;
import eu.jonahbauer.chat.bot.database.NonUniqueResultException;
import eu.jonahbauer.chat.bot.pizza.model.*;
import eu.jonahbauer.chat.bot.pizza.model.table.Order;
import eu.jonahbauer.chat.bot.pizza.model.table.OrderItem;
import eu.jonahbauer.chat.bot.pizza.model.table.Pizza;
import eu.jonahbauer.chat.bot.pizza.model.table.User;
import eu.jonahbauer.chat.bot.pizza.model.view.OrderDetail;
import eu.jonahbauer.chat.bot.pizza.model.view.OrderItemDetail;
import eu.jonahbauer.chat.bot.pizza.util.ArgumentParser;
import eu.jonahbauer.chat.bot.pizza.util.Pair;
import lombok.extern.slf4j.Slf4j;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.VisibleForTesting;

import java.sql.SQLException;
import java.time.ZoneId;
import java.time.format.DateTimeFormatter;
import java.util.*;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.stream.Collectors;

import static eu.jonahbauer.chat.bot.pizza.util.ArgumentParser.*;
import static eu.jonahbauer.chat.bot.pizza.util.FormatUtils.*;

@Slf4j
public class PizzaBot extends ChatBot {
    private static final @NotNull ScopedValue<Double> ANNOYANCE = ScopedValue.newInstance();
    private static final @NotNull ScopedValue<User> USER = ScopedValue.newInstance();

    private static final @NotNull BotConfig.Key.OfStringArray PRIMARY_CHANNELS = BotConfig.Key.ofStringArray("primary_channels");
    private static final @NotNull BotConfig.Key.OfString DATABASE = BotConfig.Key.ofString("database");

    private static final @NotNull DateTimeFormatter DATE_FORMATTER = DateTimeFormatter.ofPattern("HH:mm")
            .withZone(ZoneId.of("Europe/Berlin"))
            .withLocale(Locale.GERMANY);
    private static final @NotNull List<@NotNull String> ANNOYED_RESPONSES = List.of(
            "Es ist so dunkel, ich kann dich nicht hören.",
            "Jetzt schrei doch nicht so!",
            "Fresse!"
    );

    private final @Nullable Set<@NotNull String> primaryChannels;
    private final @NotNull PizzaService service;

    public PizzaBot() {
        super("Marek");
        this.primaryChannels = getConfig().get(PRIMARY_CHANNELS).map(Set::copyOf).orElse(null);
        this.service = new PizzaService(new Database(getConfig().require(DATABASE)));
    }

    @VisibleForTesting
    PizzaBot(@Nullable Set<@NotNull String> primaryChannels, @NotNull Database database) {
        super("Marek");
        this.primaryChannels = primaryChannels;
        this.service = new PizzaService(database);
    }

    @Override
    protected void onMessage(@NotNull Post post) {
        var message = post.message();
        if (message.length() < 6) return;

        var command = message.substring(0, 6);

        var runnable = (Runnable) () -> {
            try {
                if ("!marek".equals(command) || "!Marek".equals(command)) {
                    handle(post);
                } else if ("!MAREK".equals(command) && !ANNOYED_RESPONSES.isEmpty()) {
                    handleAnnoyed();
                } else if ("!marek".equalsIgnoreCase(command)) {
                    var uppercase = command.chars().filter(chr -> 'A' <= chr && chr <= 'Z').count();
                    ScopedValue.runWhere(ANNOYANCE, uppercase / 5d, () -> handle(post));
                }
            } catch (PizzaBotException ex) {
                post(ex.getMessage());
            }
        };

        if (post.userId() != null && post.userName() != null) {
            ScopedValue.runWhere(USER, new User(post.userId(), post.userName()), runnable);
        } else {
            runnable.run();
        }
    }

    private void handle(@NotNull Post post) {
        if (!isPrimaryChannel(post.channel())) {
            handleNonPrimaryChannel();
        } else {
            try {
                dispatch(ArgumentParser.parse(post.message().substring(6)));
            } catch (SQLException ex) {
                log.error("Fehler beim Verarbeiten von {}.", post, ex);
            }
        }
    }

    private void handleAnnoyed() {
        assert !ANNOYED_RESPONSES.isEmpty();

        var idx = (int) (Math.random() * ANNOYED_RESPONSES.size());
        post(ANNOYED_RESPONSES.get(idx));
    }

    private void handleNonPrimaryChannel() {
        assert primaryChannels != null;

        if (primaryChannels.size() == 1) {
            post(STR."Bitte benutze den Channel \{primaryChannels.iterator().next()}.");
        } else {
            var channels = new StringBuilder();
            var it = primaryChannels.iterator();
            while (it.hasNext()) {
                var channel = it.next();
                if (!it.hasNext()) {
                    channels.append(" oder ");
                } else if (!channels.isEmpty()) {
                    channels.append(", ");
                }
                channels.append(channel);
            }
            post(STR."Bitte benutze einen der Channel \{channels.toString()}.");
        }
    }

    private void dispatch(@NotNull List<@NotNull String> args) throws SQLException {
        if (args.isEmpty()) {
            helpUser();
            return;
        }

        var argc = args.size();
        var command = args.getFirst();
        switch (command) {
            case "--user", "-u" -> runAs(args);
            case "help" -> {
                if (argc > 1 && "--all".equals(args.get(1))) {
                    helpAdmin();
                } else {
                    helpUser();
                }
            }
            case "info" -> info();
            case "order" -> {
                if (argc == 1) throw new PizzaBotException("too few arguments");
                switch (args.get(1)) {
                    case "revoke" -> {
                        if (argc == 2) orderRevoke();
                        else if (argc != 3) throw new PizzaBotException("too many arguments");
                        else switch (args.get(2)) {
                            case "--all" -> orderRevokeAll();
                            case String s when isLong(s) -> orderRevoke(toLong(s));
                            case String name -> orderRevoke(name);
                        }
                    }
                    case "list" -> {
                        if (argc == 2) orderList();
//                        else if (argc == 3 && "--all".equals(args.get(2))) orderListAll();
                        else throw new PizzaBotException("invalid command");
                    }
                    case "add" -> {
                        if (argc < 4) throw new PizzaBotException("too few arguments");
                        else if (argc > 4) throw new PizzaBotException("too many arguments");
                        orderAdd(args.get(2), args.get(3));
                    }
                    case String name when argc == 2 -> order(name, null);
                    case String name -> order(name, String.join(" ", args.subList(2, args.size())));
                }
            }
            case "summary" -> {
                if (argc == 1) summary();
                else if (argc == 2 && "--all".equals(args.get(1))) summaryAll();
                else if (argc == 2) summary(args.get(1));
                else throw new PizzaBotException("invalid command");
            }
            case "check" -> {
                if (argc == 1) check();
                else if (argc == 2 && "--all".equals(args.get(1))) checkAll();
                else if (argc == 2) check(args.get(1));
                else throw new PizzaBotException("invalid command");
            }
            case "pay" -> {
                if (argc == 1) throw new PizzaBotException("missing argument");
                else if (argc == 2) pay(toDouble(args.get(1)));
                else if (argc == 3) pay(args.get(1), toDouble(args.get(2)));
                else throw new PizzaBotException("too many arguments");
            }
            case "payment" -> {
                if (argc == 1) throw new PizzaBotException("too few arguments");
                switch (args.get(1)) {
                    case "confirm" -> {
                        if (argc == 2) throw new PizzaBotException("too few arguments");
                        else if (argc == 3 && "--all".equals(args.get(2))) throw new PizzaBotException("missing arguments");
                        else if (argc == 3) paymentConfirm(toLong(args.get(2)), null);
                        else if (argc == 4 && "--all".equals(args.get(2))) paymentConfirmAll(args.get(3));
                        else if (argc == 4) paymentConfirm(toLong(args.get(2)), args.get(3));
                        else throw new PizzaBotException("invalid command");
                    }
                    case "void" -> {
                        if (argc == 2) throw new PizzaBotException("too few arguments");
                        else if (argc == 3) paymentVoid(toLong(args.get(2)), null);
                        else if (argc == 4) paymentVoid(toLong(args.get(2)), args.get(3));
                        else throw new PizzaBotException("invalid command");
                    }
                    default -> throw new PizzaBotException("invalid command");
                }
            }
            default -> throw new PizzaBotException("Unbekannter Befehl.");
        }
    }

    private @NotNull User getUser() {
        if (USER.isBound()) {
            return USER.get();
        } else {
            throw new PizzaBotException("Diese Aktion kann nur mit öffentlicher ID ausgeführt werden.");
        }
    }

    //<editor-fold desc="Admin" defaultstate="collapsed">
    private void runAs(@NotNull List<String> args) throws SQLException {
        // TODO
        if (true) throw new PizzaBotException("nicht implementiert");

        if (args.size() < 2) throw new PizzaBotException("missing argument");

        var username = args.get(1);
//        var user = service.getUserByEventAndName(, username);
//
//        ScopedValue.runWhere(USER, user, () -> {
//            try {
//                dispatch(args.subList(2, args.size()));
//            } catch (SQLException ex) {
//                throw Lombok.sneakyThrow(ex);
//            }
//        });
//
//        if (false) throw new SQLException();
    }

    private void orderAdd(@NotNull String targetOrder, @NotNull String newUserName) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var newUser = service.getUserByName(newUserName);

        OrderItem item;
        Pizza pizza;
        List<UserWithPayment> users;
        try {
            var targetId = Long.parseLong(targetOrder);
            var detail = service.getItemById(targetId)
                    .orElseThrow(() -> new PizzaBotException(STR."Die Bestellung mit Nummer \{targetId} existiert nicht."));
            item = detail.item();
            pizza = detail.pizza();
            users = detail.users();
        } catch (NumberFormatException _) {
            var user = service.getUserByEventAndName(order.eventId(), targetOrder);
            try {
                item = service.getItemByUser(order.id(), user.id())
                        .orElseThrow(() -> new PizzaBotException(STR."\{user.username()} hat keine Bestellung aufgegeben."));
                pizza = service.getPizzaById(item.pizzaId()).orElseThrow();
                users = service.getUsersByItem(item);
            } catch (NonUniqueResultException ex) {
                throw new PizzaBotException(STR."\{user.username()} hat mehr als eine Bestellung aufgegeben. Bitte gibt die Positionsnummer an.", ex);
            }
        }

        var payed = users.stream().filter(UserWithPayment::payed).map(UserWithPayment::username).toList();
        if (!payed.isEmpty()) {
            throw new PizzaBotException(STR."\{join(payed)} \{payed.size() == 1 ? "hat" : "haben"} schon bezahlt.");
        }

        if (users.stream().anyMatch(u -> u.id() == newUser.id())) {
            throw new PizzaBotException(STR."\{newUser.username()} nimmt bereits an dieser Bestellung teil.");
        }

        service.add(item, newUser, false);
        post(STR."\{newUser.username()} wurde zur Bestellung \{item.id()} (\{pizza.name()}) von \{join(users, UserWithPayment::username)} hinzugefügt.");
    }

    /**
     * Zeigt eine Übersicht alle Bestellungen während der Veranstaltung.
     */
    private void summaryAll() throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        summary0(order.eventName(), service.getItemDetailByEvent(order.eventId()));
    }

    /**
     * Zeigt eine Übersicht der Positionen der laufenden Bestellung.
     */
    private void summary() throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        summary0(order.name(), service.getItemDetailByOrder(order.id()));
    }

    /**
     * Zeigt eine Übersicht der Positionen der Bestellung mit dem angegebenen Namen.
     */
    private void summary(@NotNull String name) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var other = service.getOrderByEventAndName(order.eventId(), name);
        summary0(other.name(), service.getItemDetailByOrder(other.id()));
    }

    private void summary0(@NotNull String name, @NotNull Collection<OrderItemDetail> details) {
        var items = details.stream().collect(Collectors.groupingBy(
                detail -> Pair.of(detail.pizza(), detail.notes()),
                () -> new TreeMap<>(Comparator.comparing(Pair::first)),
                Collectors.collectingAndThen(
                        Collectors.groupingBy(
                                OrderItemDetail::id,
                                Collectors.mapping(OrderItemDetail::userName, Collectors.joining("+"))
                        ),
                        Map::values
                )
        ));
        var lines = new ArrayList<String>(items.size() + 2);
        lines.add("Übersicht " + name);
        lines.add("");
        if (items.isEmpty()) lines.add("(keine Einträge)");
        items.forEach((item, users) -> {
            var counts = users.stream().collect(Collectors.groupingBy(
                    Function.identity(),
                    Collectors.counting()
            ));
            var usersString = counts.entrySet().stream()
                    .sorted(Map.Entry.comparingByKey())
                    .map(entry -> entry.getValue() == 1 ? entry.getKey() : entry.getValue() + "x " + entry.getKey())
                    .collect(Collectors.joining(", "));
            var pizza = item.first().name().toUpperCase(Locale.ROOT);
            var notes = item.second() != null ? " " + item.second() : "";
            lines.add(STR."\{tabular(users.size())}x \{pizza}\{notes} (\{usersString})");
        });
        post(String.join("\n", lines));
    }

    /**
     * Zeigt die zusammengefasste Rechnung für alle Bestellungen während der Veranstaltung.
     */
    private void checkAll() throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var details = service.getItemDetailByEvent(order.eventId());

        var tips = details.stream().collect(Collectors.groupingBy(
                OrderItemDetail::order,
                Collectors.collectingAndThen(Collectors.toList(), list -> calculateTip(list.getFirst().orderPayment(), list))
        ));

        var items = details.stream().collect(Collectors.groupingBy(
                OrderItemDetail::user,
                Collectors.reducing(Receipt.empty(), item -> item.receipt(tips.get(item.order())), Receipt::sum)
        ));
        check0(order.eventName(), tips.values().stream().anyMatch(tip -> tip > 0), items);
    }

    /**
     * Zeigt die Rechnung für der laufenden Bestellung.
     */
    private void check() throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        check0(order.order());
    }

    /**
     * Zeigt die Rechnung für die Bestellung mit dem angegebenen Namen.
     */
    private void check(@NotNull String name) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var other = service.getOrderByEventAndName(order.eventId(), name);
        check0(other);
    }

    private static double calculateTip(@Nullable Double payment, List<@NotNull OrderItemDetail> details) {
        var total = details.stream().mapToDouble(OrderItemDetail::userPrice).sum();
        return payment == null ? 0 : payment / total;
    }

    private void check0(@NotNull Order order) throws SQLException {
        var details = service.getItemDetailByOrder(order.id());

        var tip = calculateTip(order.payment(), details);
        var items = details.stream().collect(Collectors.groupingBy(
                OrderItemDetail::user,
                Collectors.reducing(Receipt.empty(), item -> item.receipt(tip), Receipt::sum)
        ));
        check0(order.name(), order.payment() != null, items);
    }

    private void check0(@NotNull String name, boolean tipped, @NotNull Map<User, @NotNull Receipt> items) {
        var lines = new ArrayList<String>();
        lines.add("Rechnung für " + name + (tipped ? " (mit Trinkgeld)" : ""));
        lines.add("alle Angaben ohne Gewähr");
        lines.add("");
        items.forEach((user, receipt) -> {
            if (receipt.payed()) {
                lines.add(strikethrough(user.username() + ": " + receipt.totalAsString()));
            } else if (receipt.unpayed()) {
                lines.add(user.username() + ": " + receipt.totalAsString());
            } else {
                lines.add(user.username() + ": " + strikethrough(receipt.totalAsString()) + " " + receipt.pendingAsString());
            }
        });
        post(String.join("\n", lines));
    }
    
    private void pay(double amount) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        service.save(order.order().withPayment(amount));
    }
    
    private void pay(@NotNull String name, double amount) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var other = service.getOrderByEventAndName(order.eventId(), name);
        service.save(other.withPayment(amount));
    }

//    /**
//     * Beginnt eine neue Bestellung.
//     * @param restaurant die Restaurant-ID
//     */
//    private void orderStart(@NotNull String restaurant, @NotNull String deadline) {
//    }
//
//    /**
//     * Beginnt eine neue Bestellung.
//     * @param restaurantId die Restaurant-ID
//     */
//    private void orderStart(long restaurantId, @NotNull String deadline) {
//        var restaurant = restaurantRepository.getById(restaurantId)
//                .orElseThrow(() -> new BreakException(STR."Das Restaurant mit der ID \{restaurantId} existiert nicht."));
//        var order = new Order(-1, , restaurantId, , , null);
//        orderRepository.save(order);
//    }
//
//    /**
//     * Beendet eine Bestellung.
//     * @param payment der gezahlte Betrag inkl. Trinkgeld
//     */
//    private void orderFinish(double payment) {
//        var order = service.getCurrentOrderAsAdmin();
//        orderRepository.save(order.withPayment(payment));
//        check(order.id());
//    }

    /**
     * Bestätigt den Zahlungseingang für alle ausstehenden Zahlungen der laufenden Veranstaltung eines Benutzers.
     */
    private void paymentConfirmAll(@NotNull String name) throws SQLException {
        var order = service.getCurrentOrder().checkAdminAccess(getUser());
        var user = service.getUserByEventAndName(order.eventId(), name);
        var count = service.paymentByEventAndUser(order.eventId(), user.id(), true);
        post(STR."\{count} ausstehenden Zahlungen für \{user.username()} wurden bestätigt.");
    }

    /**
     * Bestätigt den Zahlungseingang für eine Bestellung.
     */
    private void paymentConfirm(long itemId, @Nullable String username) throws SQLException {
        payment0(itemId, username, true, (pizza, user) -> STR."Zahlungseingang für \{pizza} (\{user}) wurde bestätigt.");
    }

    /**
     * Widerruft die Bestätigung eines Zahlungseingangs.
     */
    private void paymentVoid(long itemId, @Nullable String username) throws SQLException {
        payment0(itemId, username, false, (pizza, user) -> STR."Zahlungseingang für \{pizza} (\{user}) wurde bestätigt.");
    }

    private void payment0(long itemId, @Nullable String username, boolean payed, @NotNull BinaryOperator<String> message) throws SQLException {
        var _ = service.getCurrentOrder().checkAdminAccess(getUser());
        var summary = service.getItemById(itemId)
                .orElseThrow(() -> new PizzaBotException(STR."Die Bestellung mit Nummer \{itemId} existiert nicht."));
        if (username != null) {
            var users = summary.users().stream().filter(u -> u.username().startsWith(username)).toList();
            if (users.isEmpty()) {
                throw new PizzaBotException("Der Benutzer existiert nicht oder hat nicht an der Bestellung teilgenommen.");
            } else if (users.size() > 1) {
                throw new PizzaBotException("Der Benutzer ist nicht eindeutig.");
            } else {
                var user = users.getFirst();
                service.paymentByItemAndUser(itemId, user.id(), true);
                post(message.apply(summary.pizza().name(), user.username()));
            }
        } else {
            service.paymentByItem(summary.item().id(), payed);
            post(message.apply(summary.pizza().name(), join(" + ", summary.users(), UserWithPayment::username)));
        }
    }
    //</editor-fold>

    //<editor-fold desc="User" defaultstate="collapsed">
    /**
     * Zeigt Informationen zur aktuell laufenden Bestellung.
     */
    private void info() throws SQLException {
        var order = service.getCurrentOrder();
        var result = service.getRestaurantWithPizzas(order.restaurantId());
        var restaurant = result.first();
        var pizzas = result.second();

        var lines = new ArrayList<String>();
        lines.add(STR."Pizzabestellung bei \{restaurant.name()} in \{restaurant.city()}");
        lines.add(STR."Bestellannahmeschluss: \{DATE_FORMATTER.format(order.deadline())}");
        pizzas.stream().map(Pizza::toPrettyString).forEach(lines::add);
        post(String.join("\n", lines));
    }

    /**
     * Listet die eigenen Bestellungen auf.
     */
    private void orderList() throws SQLException {
        var user = getUser();
        var current = service.getCurrentOrder();

        var itemsByOrder = service.getItemDetailByEventAndUser(current.eventId(), user.id()).stream().collect(Collectors.groupingBy(
                OrderItemDetail::order,
                TreeMap::new,
                Collectors.toUnmodifiableList()
        ));

        var lines = new ArrayList<String>();
        lines.add("Bestellungen von " + user.username());
        if (itemsByOrder.isEmpty()) lines.add("(keine Einträge)");
        itemsByOrder.forEach((order, items) -> {
            lines.add("");
            lines.add(order.name());
            items.stream().map(OrderItemDetail::toPrettyString).forEach(lines::add);
        });
        post(String.join("\n", lines));
    }

    /**
     * Storniert die einzige eigene Bestellung.
     */
    private void orderRevoke() throws SQLException {
        var order = service.getCurrentOrder().checkDeadline();
        var user = getUser();

        try {
            var item = service.getItemByUser(order.id(), user.id());
            if (item.isEmpty()) {
                post("Du hast keine Bestellung aufgegeben.");
            } else {
                orderRevoke0(item.get());
            }
        } catch (NonUniqueResultException ex) {
            post("Du hast mehr als eine Bestellung aufgegeben. Bitte gib die Positionsnummer oder den Namen der Pizza an.");
        }
    }

    /**
     * Storniert die angegebene, eigene Bestellung.
     */
    private void orderRevoke(long itemId) throws SQLException {
        var order = service.getCurrentOrder().checkDeadline();
        var item = service.getItemById(itemId);
        if (item.isEmpty() || item.get().item().orderId() != order.id()) {
            post(STR."Die Position mit Nummer \{itemId} existiert nicht.");
        } else {
            orderRevoke0(item.get().checkUserAccess(getUser()).item());
        }
    }

    /**
     * Storniert die eigene Bestellung der angegebene Pizza.
     */
    private void orderRevoke(@NotNull String name) throws SQLException {
        var order = service.getCurrentOrder().checkDeadline();
        var user = getUser();
        try {
            var item = service.getItemByPizzaName(order.id(), user.id(), name);
            if (item.isEmpty()) {
                post("Du hast keine solche Pizza bestellt.");
            } else {
                orderRevoke0(item.get());
            }
        } catch (NonUniqueResultException ex) {
            post("Du hast mehr als eine solche Pizza bestellt. Bitte gib die Positionsnummer an.");
        }
    }

    private void orderRevoke0(@NotNull OrderItem item) throws SQLException {
        var pizza = service.getPizzaById(item.pizzaId()).orElseThrow();
        service.delete(item);
        post(STR."Die Position \{item.id()} (\{pizza.name()}) wurde storniert.");
    }

    /**
     * Storniert alle eigenen Bestellungen.
     */
    private void orderRevokeAll() throws SQLException {
        var order = service.getCurrentOrder().checkDeadline();
        var user = getUser();
        var items = service.getItemsByUser(order.id(), user.id());
        if (items.isEmpty()) {
            post("Du hast keine Bestellung aufgegeben.");
        } else {
            for (var item : items) {
                service.delete(item);
            }
            post("Alle deine Bestellungen wurden storniert.");
        }
    }

    /**
     * Bestellt die Pizza mit dem angegebenen Namen.
     */
    private void order(@NotNull String nameOrNummer, @Nullable String notes) throws SQLException {
        var order = service.getCurrentOrder().checkDeadline();
        var pizza = service.getPizza(order.restaurantId(), nameOrNummer);
        order0(order, pizza, notes);
    }

    private void order0(@NotNull OrderDetail order, @NotNull Pizza pizza, @Nullable String notes) throws SQLException {
        var id = service.order(order, getUser(), pizza, notes);
        post(STR."Pizza \{pizza.name()} wurde zur Bestellung hinzugefügt (Positionsnummer \{id}).");
    }
    //</editor-fold>

    private void helpUser() {
        post("""
             Marek is back!
             !marek info - Menü anzeigen
             
             Bestellung
             !marek order (NUMMER | NAME) [NOTIZEN] - Pizza bestellen
             !marek order revoke [NUMMER | NAME | --all] - Bestellung(en) stornieren
             !marek order list - Bestellungen auflisten
             """);
    }

    private void helpAdmin() {
        post("""
             !marek --user USER order (NUMMER | NAME) [NOTIZEN] - Pizza bestellen
             !marek --user USER order revoke [NUMMER | NAME | --all] - Bestellung(en) stornieren
             !marek --user USER order list - Bestellungen auflisten
             
             !marek summary [NAME | --all] - Bestellübersicht
             !marek check [NAME | --all] - Rechnung anzeigen
             !marek pay [NAME] AMOUNT - Rechnungsbetrag angeben
             !marek order list --all - Alle Bestellungen auflisten
             !marek order add (NUMMER | NAME) USER - Fügt USER zur Bestellung mit Nummer NUMMER / von NAME hinzu
             
             !marek payment confirm NUMMER [USER] - Zahlungseingang bestätigen
             !marek payment confirm --all USER - Zahlungseingang bestätigen
             !marek payment void NUMMER [USER] - Widerruf der Bestätigung des Zahlungseingangs
             """);
    }

    @Override
    protected void post(@NotNull String message, @Nullable String name, @Nullable Boolean bottag, @Nullable Boolean publicId) {
        var annoyance = ANNOYANCE.orElse(0d);
        if (annoyance == 0) {
            super.post(message, name, bottag, publicId);
        } else if (annoyance == 1) {
            super.post(message.toUpperCase(Locale.ROOT), name, bottag, publicId);
        } else {
            var infuriatedMessage = message.codePoints()
                    .map(chr -> Math.random() < annoyance ? Character.toUpperCase(chr) : chr)
                    .collect(StringBuilder::new, StringBuilder::appendCodePoint, StringBuilder::append)
                    .toString();
            super.post(infuriatedMessage, name, bottag, publicId);
        }
    }

    private boolean isPrimaryChannel(@NotNull String channel) {
        return primaryChannels == null || primaryChannels.contains(channel);
    }
}
